/*
    GPSUTIL a program to interact with NMEA compatable and other supported
      GPS units.

    Copyright (C) 2000  Brian J. Hennings

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    I can be contacted at hennings@cs.uakron.edu or by US mail at:

    Brian Hennings
    114 Rex Ave.
    Wintersville, OH 43953
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#include <windows.h>
#endif

#include "gps.h"
#include "serial.h"
#include "magellan.h"

extern CPORT Serial;
extern int   debug;

/***********************/
/* Function Prototypes */
/***********************/
unsigned char MChecksum (char *Data);

/*****************************************************************************
 * MChecksum () - Takes a complete Megellan message and computes the
 *     checksum value.  **NOTE**  The Protocol document v1.0 is incorrect
 *     about the checksum value -- it DOES NOT include the $ or the *
 *****************************************************************************/
unsigned char MChecksum (char *Data)
{unsigned char CheckVal;
          int  len, idx;

 CheckVal = 0;
 len = strlen (Data) - 1;  //Don't include the '*' at end of message!

//idx starts at 1 becuase the '$' isn't part of the checksum!
 for (idx = 1; idx < len; idx++) {
   CheckVal = CheckVal ^ Data[idx];   //The ^ is an XOR in C
 } /*End of for*/

 return CheckVal;
} /*End of MChecksum ()*/

/*****************************************************************************
 * AddWpt () - Adds a waypoint to a linked list.  Note -- due to the fun
 *     involved with double pointers, I decided for easier to follow code to
 *     expect an empty list to contain 1 element with the name string NULL.
 *****************************************************************************/
static void AddWpt (char *Msg, MWpt *List)
{MWpt *Node = NULL, *At = NULL;
 char *Cur, Tmp[10];
 int Len, idx;

 Node = (MWpt *) malloc (sizeof (MWpt));
 if (Node == NULL) {
   printf ("Panic!!!  malloc (MWPT) failed!\n");
   exit (-99);
 }

 Len = strlen (Msg);
 Cur = Msg;
 Cur = strchr (Cur, ',');
 Cur++;

 for (idx = 0; idx < 8; idx++)
   Tmp[idx] = Cur[idx];
 Tmp[8] = '\0';
 Node->Latitude = atof (Tmp);
 Cur += 9;
 Node->LatDir = Cur[0];
 Cur += 2;
 for (idx = 0; idx < 9; idx++)
   Tmp[idx] = Cur[idx];
 Node->Longitude = atof (Tmp);
 Cur += 10;
 Node->LongDir = Cur[0];
 Cur += 2;
 for (idx = 0; idx < 7; idx++)
   Tmp[idx] = Cur[idx];
 Tmp[7] = '\0';
 Node->Altitude = atol (Tmp);
 Cur += 8;
 Node->AltType = Cur[0];
 Cur += 2;
 for (idx = 0; idx < MNAME_SIZE; idx++)
   if (Cur[idx] == ',')
     break;
   else
     Node->Name[idx] = Cur[idx];

 Node->Name[idx] = '\0';
 Cur += idx + 1;
 for (idx = 0; idx < MDESC_SIZE; idx++)
   if (Cur[idx] == ',')
     break;
   else
     Node->Desc[idx] = Cur[idx];

 Node->Desc[idx] = '\0';
 Cur += idx + 1;
 Node->Icon = Cur[0];
 Node->Next = NULL;

 if (List->Name[0] == '\0') {
   memcpy (List, Node, sizeof (MWpt));
   free (Node);
 } /*End of if*/
 else {
   At = List;
   while (At->Next != NULL)
     At = At->Next;
   At->Next = Node;
 } /*End of else*/

} /*End of AddWpt ()*/

/*****************************************************************************
 * MagellanULWPt () - Sends a single waypoint to the GPS receiver.
 *****************************************************************************/
int MagellanULWPt (MWpt Point)
{char Tmp[MAXLEN], Msg[MAXLEN];
 unsigned char Check;
 int RetVal = 0;

 strcpy (Tmp, "$PMGNCMD,HANDON*");
 Check = MChecksum (Tmp);
 sprintf (Msg, "%s%02X", Tmp, Check);
 if (debug > 0)
   printf ("Sending: %s\n", Msg);
 WriteMessage (Msg);

 sprintf (Tmp, "$PMGNWPL,%04.3f,%c,%05.3f,%c,%07ld,%c,%s,%s,%c*",
          Point.Latitude, Point.LatDir, Point.Longitude, Point.LongDir,
          Point.Altitude, Point.AltType, Point.Name, Point.Desc,
          Point.Icon);
 Check = MChecksum (Tmp);
 sprintf (Msg, "%s%02X", Tmp, Check);
 if (debug > 0)
   printf ("Sending: %s\n", Msg);
 WriteMessage (Msg);
 strcpy (Tmp, "$PMGNCMD,HANDOFF*");
 Check = MChecksum (Tmp);
 sprintf (Msg, "%s%02X", Tmp, Check);
 if (debug > 0)
   printf ("Sending: %s\n", Msg);
 WriteMessage (Msg);

 return RetVal;
} /*End of MagellanULWPt ()*/

/*****************************************************************************
 * MagellanDLWPts () - Creates a linked list of all waypoints stored on the
 *     GPS receiver.
 *****************************************************************************/
int MagellanDLWPts (MWpt *List)
{char Tmp[MAXLEN], Msg[MAXLEN], Old[MAXLEN], CVal[3];
 unsigned char Check;
 int len, Done = FALSE, ECnt, rc, RetVal = 0;


 CVal[2] = '\0';
 ECnt = 0;
 strcpy (Tmp, "$PMGNCMD,HANDON*");
 Check = MChecksum (Tmp);
 sprintf (Msg, "%s%02X", Tmp, Check);
 if (debug > 0)
   printf ("Sending: %s\n", Msg);
 WriteMessage (Msg);
 strcpy (Tmp, "$PMGNCMD,WAYPOINT*");
 Check = MChecksum (Tmp);
 sprintf (Msg, "%s%02X", Tmp, Check);
 if (debug > 0)
   printf ("Sending: %s\n", Msg);
 WriteMessage (Msg);
 while (!Done) {
   rc = FindMessage ("$PMGN", Msg, MAXLEN);
   if (rc == 0) {
     ECnt = 0;
     if (debug > 0)
       printf ("Received: %s\n", Msg);
     strcpy (Old, Msg);
     if ((Msg[5] == 'C') && (Msg[6] == 'S') && (Msg[7] == 'M'))
       continue;
     if ((Msg[5] == 'W') && (Msg[6] == 'P') && (Msg[7] == 'L'))
       AddWpt (Msg, List);
     len = strlen (Msg);
     CVal[0] = Msg[len - 2];
     CVal[1] = Msg[len - 1];
     sprintf (Tmp, "$PMGNCSM,%s*", CVal);
     Check = MChecksum (Tmp);
     sprintf (Msg, "%s%02X", Tmp, Check);
     if (debug > 0)
       printf ("Sending: %s\n", Msg);
     WriteMessage (Msg);
     if (strcmp (Old, "$PMGNCMD,END*3D") == 0) {
       Done = TRUE;
     }
   } /*End of if*/
   else {
     ECnt++;
     if (ECnt > MAXERR) {
       Done = TRUE;
       RetVal = -1;
     } /*End of if*/
   } /*End of else*/
 } /*End of while*/

 strcpy (Tmp, "$PMGNCMD,HANDOFF*");
 Check = MChecksum (Tmp);
 sprintf (Msg, "%s%02X", Tmp, Check);
 if (debug > 0)
   printf ("Sending: %s\n", Msg);
 WriteMessage (Msg);

 return RetVal;
} /*End of MAgellanDLWPts ()*/

/*****************************************************************************
 * MDumpWPts () - Writes out the linked list of waypoints to a file in a fixed
 *     format.
 *****************************************************************************/
void MDumpWPts (char *FName, MWpt *List)
{MWpt *Cur;
 FILE *OFile;

 if (List->Name[0] == '\0')
   return;

 if (strcmp (FName, "stdout") == 0)
   OFile = stdout;
 else {
   if ((OFile = fopen (FName, "w")) == NULL) {
     printf ("Unable to open output file!\n");
     return;
   } /*End of if*/
 } /*End of else*/

 Cur = List;
 fprintf (OFile, "#NAME  LATITUDE  LONGITUDE  ALTITUDE Description         Icon\n");

 while (Cur != NULL) {
   fprintf (OFile, "%-6s %08.3f%c %09.3f%c %07ld%c %-20s %c\n",
            Cur->Name, Cur->Latitude, Cur->LatDir, Cur->Longitude,
            Cur->LongDir, Cur->Altitude, Cur->AltType, Cur->Desc,
            Cur->Icon);
   Cur = Cur->Next;
 } /*End of while*/
} /*End of DumpWPts ()*/

/*****************************************************************************
 * FixStr () - A quick function to remove all trailing spaces at the end of a
 *     string.
 *****************************************************************************/
static void FixStr (char *Str)
{int Len, idx;

 Len = strlen (Str) - 1;
 for (idx = Len; idx >= 0; idx--)
   if (Str[idx] == ' ')
     Str[idx] = '\0';
   else
     break;

} /*End of FixStr ()*/

/*****************************************************************************
 * RestoreWPts () - Sends all waypoints in the file to the GPS unit.
 *****************************************************************************/
void MRestoreWPts (char *FName)
{MWpt Coord;
 int  idx, Done = FALSE;
 char InChar, Tmp[25];
 FILE *IFile;

 if ((IFile = fopen (FName, "r")) == NULL) {
   printf ("Unable to open input file!\n");
   return;
 } /*End of if*/

 while (!Done) {
   InChar = fgetc (IFile);
   if (InChar == EOF)
     break;
   if (InChar == '#') {
     while (fgetc (IFile) != '\n');  //Lines that start with # are skipped!
   } /*End of if*/
   else
     ungetc (InChar, IFile);

   for (idx = 0; idx < 6; idx++)
     Tmp[idx] = fgetc (IFile);
   Tmp[idx] = '\0';
   FixStr (Tmp);
   strcpy (Coord.Name, Tmp);
   fgetc (IFile);                    //Space between each column.
   for (idx = 0; idx < 8; idx++)
     Tmp[idx] = fgetc (IFile);
   Tmp[idx] = '\0';
   Coord.Latitude = atof (Tmp);
   Coord.LatDir = fgetc (IFile);
   fgetc (IFile);                    //Space between each column.
   for (idx = 0; idx < 9; idx++)
     Tmp[idx] = fgetc (IFile);
   Tmp[idx] = '\0';
   Coord.Longitude = atof (Tmp);
   Coord.LongDir = fgetc (IFile);
   fgetc (IFile);                    //Space between each column.
   for (idx = 0; idx < 7; idx++)
     Tmp[idx] = fgetc (IFile);
   Tmp[idx] = '\0';
   Coord.Altitude = atol (Tmp);
   Coord.AltType = fgetc (IFile);
   fgetc (IFile);                    //Space between each column.
   for (idx = 0; idx < 20; idx++)
     Tmp[idx] = fgetc (IFile);
   Tmp[idx] = '\0';
   FixStr (Tmp);
   strcpy (Coord.Desc, Tmp);
   fgetc (IFile);                    //Space between each column.
   Coord.Icon = fgetc (IFile);
   while (fgetc (IFile) != '\n');    //Skip rest of line.

   MagellanULWPt (Coord);
 } /*End of while*/

} /*End of RestoreWPts ()*/

/*****************************************************************************
 * MagellanInit () - Used to determine that a Megellan GPS is attached to the
 *    specified port.
 *****************************************************************************/
int MagellanInit ()
{char Tmp[MAXLEN], Msg[MAXLEN], Prefix[10];
 unsigned char Check;
 int RVal;

 strcpy (Tmp, "$PMGNCMD,VERSION*");
 strcpy (Prefix, "$PMGNVER");
 Check = MChecksum (Tmp);
 sprintf (Msg, "%s%02X", Tmp, Check);
 if (debug > 0)
   printf ("Sending: %s\n", Msg);
 WriteMessage (Msg);
 RVal = FindMessage (Prefix, Msg, MAXLEN);

 if ((RVal == 0) && (debug > 0))
   printf ("Found: %s\n", Msg);

 return 0;
} /*End of MagellanInit ()*/

/*****************************************************************************
 * MFreeWPts () - Simple function to free up memory allocated during building
 *     a waypoint list.
 *****************************************************************************/
int MFreeWPts (MWpt *List)
{MWpt *Cur, *Tmp;

 Cur = List->Next;
 while (Cur != NULL) {
   Tmp = Cur;
   Cur = Cur->Next;
   free (Tmp);
 } /*End of while*/

 return 0;
} /*End of MFreeWPts ()*/

